
import re
import typing


__author__ = "Jérémie Lumbroso <lumbroso@cs.princeton.edu>"

__all__ = [
    "ImslpSearchExpression",
    "ImslpSearchExpressionSubpart",
    "check_search_expr_to_query",
]


ImslpSearchExpressionSubpart = typing.Union[
    str,
    typing.Pattern[str],
    typing.Callable[[str], bool],

]

ImslpSearchExpression = typing.Optional[typing.Union[
    ImslpSearchExpressionSubpart,
    typing.List[ImslpSearchExpressionSubpart],
]]


def check_search_expr_to_query(
        query: str,
        search_expr: ImslpSearchExpression = None,
        case_insensitive: bool = True,
        intersect=True,
):
    if search_expr is None:
        return True

    if isinstance(search_expr, str):
        if case_insensitive:
            query = query.lower()
            search_expr = search_expr.lower()
        return search_expr in query

    elif isinstance(search_expr, typing.Pattern):
        m = search_expr.search(query)
        return m is not None

    elif isinstance(search_expr, typing.Callable):
        return search_expr(query)

    elif isinstance(search_expr, typing.List):

        for search_expr_item in search_expr:
            rec_call = check_search_expr_to_query(
                query=query,
                search_expr=search_expr_item,
                case_insensitive=case_insensitive,
                intersect=intersect,
            )
            if intersect:
                if not rec_call:
                    return False
            else:
                if rec_call:
                    return True

        return intersect

    return False




__author__ = "Jérémie Lumbroso <lumbroso@cs.princeton.edu>"

__all__ = [
    "IMSLP_SUBCATEGORIES",
    "IMSLP_COOKIE_NAME_USERNAME",
]


# These are the subcategories available for some categories
# NOTE: should this be an enum?
IMSLP_SUBCATEGORIES = [
    "Compositions",
    "Collaborations",
    "Collections",

    "As Arranger",
    "As Copyist",
    "As Dedicatee",
    "As Editor",

    "Books",
    "Pasticcios",
]

# IMSLP cookie name where user is stored
IMSLP_COOKIE_NAME_USERNAME = "imslp_wikiUserName"




import typing

import mwclient

import imslp
import imslp.helpers
import imslp.helpers.string_search
import imslp.interfaces
import imslp.interfaces.constants
import imslp.interfaces.internal
import imslp.interfaces.mw_api
import imslp.interfaces.scraping


__author__ = "Jérémie Lumbroso <lumbroso@cs.princeton.edu>"

__all__ = [
    "ImslpClient",
]


class ImslpClient:

    # the mwclient object
    _site = None               # type: mwclient.Site

    # the download wait time (in seconds)
    _wait_time = 15            # type: int

    def __init__(self, username=None, password=None):

        self._site = imslp.interfaces.mw_api.ImslpMwClient(
            username=username,
            password=password,
        )

    def login(
            self,
            username: str = None,
            password: str = None,
            cookies: dict = None,
            domain: str = None,
    ):
        self._site.login(
            username=username,
            password=username,
            cookies=cookies,
            domain=domain,
        )

    @property
    def logged_in(self) -> bool:
        """
        Returns whether the client is currently logged in.
        """
        return (
                self._site is not None and
                self._site.connection is not None and
                self._site.connection.cookies is not None and
                len(self._site.connection.cookies) > 1 and
                imslp.interfaces.constants.IMSLP_COOKIE_NAME_USERNAME in self._site.connection.cookies and
                self._site.connection.cookies[imslp.interfaces.constants.IMSLP_COOKIE_NAME_USERNAME] != ""
        )

    @property
    def username(self) -> typing.Optional[str]:
        """
        Returns the username of the logged in account if the client is
        logged in, and `None` otherwise.
        """
        if not self.logged_in:
            return

        return self._site.connection.cookies[imslp.interfaces.constants.IMSLP_COOKIE_NAME_USERNAME]

    @staticmethod
    def search_works(
        title: imslp.helpers.string_search.ImslpSearchExpression = None,
        composer: imslp.helpers.string_search.ImslpSearchExpression = None,
        intersect: bool = True,
        case_insensitive: bool = True
    ) -> typing.Set[imslp.interfaces.internal.HashablePageRecord]:
        """

        :param title:
        :param composer:
        :param intersect:
        :param case_insensitive:
        :return:
        """

        results = set(imslp.interfaces.internal.list_works())

        results_by_title = results
        if title is not None:
            results_by_title = set(filter(
                lambda record:
                imslp.helpers.string_search.check_search_expr_to_query(
                    query=record["intvals"]["worktitle"],
                    search_expr=title,
                    intersect=intersect,
                    case_insensitive=case_insensitive),
                results_by_title
            ))

        results_by_composer = results
        if composer is not None:
            results_by_composer = set(filter(
                lambda record:
                imslp.helpers.string_search.check_search_expr_to_query(
                    query=record["intvals"]["composer"],
                    search_expr=composer,
                    intersect=intersect,
                    case_insensitive=case_insensitive),
                results_by_composer
            ))

        if intersect:
            results = results_by_title.intersection(results_by_composer)
        else:
            results = results_by_title.union(results_by_composer)

        return results

    @staticmethod
    def search_people(
            name: imslp.helpers.string_search.ImslpSearchExpression,
            intersect: bool = True,
            case_insensitive: bool = True
    ) -> typing.Set[imslp.interfaces.internal.HashablePageRecord]:
        """

        :param name:
        :param intersect:
        :param case_insensitive:
        :return:
        """

        results = set(imslp.interfaces.internal.list_people())

        if name is not None:
            results = set(filter(
                lambda record:
                imslp.helpers.string_search.check_search_expr_to_query(
                    query=record["id"],
                    search_expr=name,
                    intersect=intersect,
                    case_insensitive=case_insensitive),
                results
            ))

        return results





import mwclient


__author__ = "Jérémie Lumbroso <lumbroso@cs.princeton.edu>"

__all__ = [
    "ImslpMwClient",
]


#
IMSLP_HOST = "imslp.org"
IMSLP_API_PATH = "/"
IMSLP_SCHEME = "https"


class ImslpMwClient(mwclient.Site):

    _site = None

    def __init__(self, username=None, password=None):
        super().__init__(
            host=IMSLP_HOST,
            path=IMSLP_API_PATH,
            scheme=IMSLP_SCHEME,
        )

        if username is not None and password is not None:
            self.login(username=username, password=password)

    def login(
        self,
        username: str = None,
        password: str = None,
        cookies: dict = None,
        domain: str = None,
    ):
        # these cookies are necessary to allow for proper download of images/PDFs
        combined_cookies = {
            "imslp_wikiLanguageSelectorLanguage": "en",
            "imslpdisclaimeraccepted": "yes",
        }

        # include user provided cookies
        if cookies is not None:
            combined_cookies.update(cookies)

        super().login(
            username=username,
            password=password,
            cookies=combined_cookies,
            domain=domain,
        )




import glob
import json
import os
import typing
import zipfile

import requests

import imslp


__author__ = "Jérémie Lumbroso <lumbroso@cs.princeton.edu>"

__all__ = [
    "HashablePageRecord",

    "list_people",
    "list_works",
    "load_cache",
]


# URL patterns for the IMSLP API
# For more information, see: https://imslp.org/wiki/IMSLP:API

# Generic URL pattern
IMSLP_API_GENERIC = (
    "http://imslp.org/imslpscripts/API.ISCR.php?"
    "account=worklist/disclaimer=accepted/sort=id/type={typ}/start={{start}}/retformat=json")

# URL for list of people (composers, performers, editors, etc.)
IMSLP_API_PEOPLE = IMSLP_API_GENERIC.format(typ=1)

# URL for list of works
IMSLP_API_WORKS = IMSLP_API_GENERIC.format(typ=2)


# Internal variables to cache calls
_cache_people = None
_cache_works = None


class HashablePageRecord(dict):
    """
    Subclass of a dictionary that can contain an IMSLP API record that
    can be hashed, using a string hash of the `permlink` of the record.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __hash__(self):
        hash_id = hash(self["permlink"])
        return hash_id


def _raw_call(imslp_url_pattern: str, start: int = 0, count: int = None):
    """
    Returns the result of a request made to the internal IMSLP API
    to retrieve a list of works or of people. This method is a helper
    method called by `list_people()` and `list_works()`.
    """

    results = []
    more_results = True

    while more_results:
        try:
            req = requests.get(imslp_url_pattern.format(start=start))
            if not req.ok:
                continue
            obj = req.json()
        except:
            break

        more_results = False

        if "metadata" in obj:
            metadata = obj.get("metadata", dict())
            more_results = metadata.get("moreresultsavailable", False)
            del obj["metadata"]

        new_results = list(obj.values())
        start += len(new_results)
        results += new_results

        if count is not None and 0 <= count <= len(results):
            return results[:count]

    return results[:count]


def list_people(
        start: int = 0,
        count: typing.Optional[int] = None,
        cache: bool = True
) -> typing.List[HashablePageRecord]:
    if cache:
        if _cache_people is None:
            load_cache()
        return _cache_people[start:start+count if count is not None else None]

    return list(map(HashablePageRecord, _raw_call(
        imslp_url_pattern=IMSLP_API_PEOPLE,
        start=start,
        count=count,
    )))


def list_works(start: int = 0, count: int = None, cache: bool = True):
    if cache:
        if _cache_works is None:
            load_cache()
        return _cache_works[start:start+count if count is not None else None]

    return list(map(HashablePageRecord, _raw_call(
        imslp_url_pattern=IMSLP_API_WORKS,
        start=start,
        count=count,
    )))


def reset_cache(from_file: bool = True) -> typing.NoReturn:
    """

    :param from_file:
    :return:
    """

    global _cache_people, _cache_works

    _cache_people = None
    _cache_works = None

    return load_cache(from_file=from_file)


# noinspection PyBroadException
def load_cache(from_file: bool = True) -> typing.NoReturn:
    """
    Loads a copy of the IMSLP people and works to memory cache, so that
    subsequent queries will be faster; if `from_file` is `True`, this
    method will attempt to load a partial cache from the package's
    internal files and only request newer entries from IMSLP. This
    considerably speeds up operations by about 4 minutes on a typical
    high-speed bandwidth.

    :param from_file: Determines whether to use package's internal cache
    :return:
    """

    global _cache_people, _cache_works

    # this is slow when not using the disk cache (which makes the
    # package quite large unfortunately! but at least not routinely
    # downloading 60 MB)
    #
    # on my computer + with my bandwidth (just as order of magnitude):
    # - list_people: 164880 items, 200 seconds to load
    # - list_works: 30559 items, 35 seconds to load

    if _cache_people is None:

        if from_file:
            # Try to load from file
            possible_files = sorted(
                list(map(os.path.abspath,
                         map(os.path.expanduser,
                             glob.glob(os.path.join(
                                 imslp.__path__[0],
                                 "../cache/imslp-people-cache-*.zip"))))),
                reverse=True)

            if len(possible_files) > 0:
                zf = zipfile.ZipFile(possible_files[0], mode="r")
                partial_contents = json.loads(zf.read(zf.namelist()[0]))
                try:
                    remainder_contents = list_people(start=len(partial_contents), cache=False)
                    _cache_people = partial_contents + remainder_contents
                except:
                    _cache_people = None

                _cache_people = list(map(HashablePageRecord, _cache_people))

        if _cache_people is None:
            _cache_people = list_people(cache=False)

    if _cache_works is None:

        if from_file:
            # Try to load from file
            possible_files = sorted(
                list(map(os.path.abspath,
                         map(os.path.expanduser,
                             glob.glob(os.path.join(
                                 imslp.__path__[0],
                                 "../cache/imslp-works-cache-*.zip"))))),
                reverse=True)

            if len(possible_files) > 0:
                zf = zipfile.ZipFile(possible_files[0], mode="r")
                partial_contents = json.loads(zf.read(zf.namelist()[0]))
                try:
                    remainder_contents = list_works(start=len(partial_contents), cache=False)
                    _cache_works = partial_contents + remainder_contents
                except:
                    _cache_works = None

                _cache_works = list(map(HashablePageRecord, _cache_works))

        if _cache_works is None:
            _cache_works = list_works(cache=False)





import json
import re
import typing
import urllib.parse

import bs4
import mwclient.page
import requests


__author__ = "Jérémie Lumbroso <lumbroso@cs.princeton.edu>"

__all__ = [
    "fetch_category_table",
    "fetch_images_metadata",
]


# Pattern of the URL base
IMSLP_BASE_URL = "https://imslp.org{}"

# Regular expression to extract the ratings
IMSLP_REGEXP_RATINGS = re.compile(r"IMSLPRatings=({[^}]+})")

# Regular expression to extract the page count
IMSLP_REGEXP_PAGE_COUNT = re.compile(r"(\d+)\s*pp*\.*")

# Pattern of URL to fetch category charts
IMSLP_SCRAPE_CATCHART_URL = IMSLP_BASE_URL.format("/index.php?title={}&customcat=ccperson1")

# class name and content string used to locate chart of works
IMSLP_SCRAPE_CATCHART_TABLE_CLASS = "wikitable"
IMSLP_SCRAPE_CATCHART_NEXT_CLASS = "next 200"
IMSLP_SCRAPE_CATCHART_NEXT_TEXT = "next 200"


def _extract_tag_text(tag: bs4.element.Tag) -> str:
    """
    Extracts the text content from a BeautifulSoup `Tag` object.

    :param tag: A BeautifulSoup tag.
    :return: The text contents of the tag.
    """
    if tag is None or tag.text is None:
        return ""

    # Convert non-breakable spaces to breakable spaces and strip ends
    return tag.text.replace("\xa0", " ").strip()


def fetch_category_table(category_name: str, subcategory: typing.Optional[str] = None):
    """
    Fetches the chart containing an index to a category's works.

    :param category_name:
    :param subcategory:
    :return:
    """

    # Compute the category name
    catname_full = "Category:{}".format(category_name)
    if subcategory is not None:
        catname_full += "/"
        catname_full += subcategory
    catname_urlenc = urllib.parse.quote(catname_full, safe='/:')

    # Accumulator variables
    header = None
    rows_parsed = []

    next_page_link = IMSLP_SCRAPE_CATCHART_URL.format(catname_urlenc)

    while next_page_link is not None:

        # Fetch the data chunk
        r = requests.get(next_page_link)
        s = bs4.BeautifulSoup(r.content)
        ts = s.find_all(name="table", attrs={"class": IMSLP_SCRAPE_CATCHART_TABLE_CLASS})
        t = ts[0]

        # Check whether there is more to collect
        next_page_link = None
        next_page_link_candidates = s.find_all(
            "a",
            attrs={
                "title": catname_full,
                "class": IMSLP_SCRAPE_CATCHART_NEXT_CLASS},
            text=IMSLP_SCRAPE_CATCHART_NEXT_TEXT,
        )
        if next_page_link_candidates is not None and len(next_page_link_candidates) > 0:
            next_page_link = IMSLP_BASE_URL.format(next_page_link_candidates[0]["href"])

        # Parse the data
        if header is None:
            # parse new header
            header = list(map(lambda tag: tag.text.strip(), t.find_all("th")))
        else:
            # confirm header is the same
            assert header == list(map(lambda tag: tag.text.strip(), t.find_all("th")))

        rows = t.find_all("tr")[1:]

        # Add to already parsed rows
        rows_parsed += [list(map(_extract_tag_text, row.find_all("td"))) for row in rows]

    # Turn into dicts
    rows_as_dicts = list(map(lambda parsed_row: dict(zip(header, parsed_row)), rows_parsed))

    return rows_as_dicts


def fetch_images_metadata(page: mwclient.page.Page) -> list:
    """
    Fetches the metadata associated with the images of an IMSLP page, as
    specified by a `mwclient.page.Page` object. This contains the download
    counter which is not available through the MediaWiki API and requires
    scraping to obtain.

    :param page:
    :return:
    """

    if page is None:
        return list()

    esc_title = urllib.parse.quote(page.base_title.replace(" ", "_"))

    u = "https://imslp.org/wiki/{}".format(esc_title)

    r = requests.get(u)
    if not r.ok:
        return list()

    s = bs4.BeautifulSoup(r.content, features="html.parser")

    images = []

    # obtain ratings dictionary from embedded JavaScript
    # (the ratings are then loaded into the DOM onLoad)
    ratings_dict = dict()
    m = IMSLP_REGEXP_RATINGS.search(s.__str__())
    if m is not None:
        ratings_dict_str = m.group(1)
        ratings_dict_old = json.loads(ratings_dict_str)

        for key, value in ratings_dict_old.items():
            try:
                ratings_dict[int(key)] = {
                    "rating": value[0],
                    "count": int(value[1]),
                }
            except ValueError:
                continue

    for f in page.images():

        f_title = f.base_title
        f_esc_title = urllib.parse.quote(f_title.replace(" ", "_"))

        # Hacky way of finding the relevant metadata
        t1 = s.find(attrs={"href": "/wiki/File:{}".format(f_esc_title)})
        t2 = s.find(attrs={"title": "File:{}".format(f_title)})

        if t1 is None and t2 is None:
            continue

        t = t1 or t2
        if t.text.strip() == "":
            continue

        page_count = None
        m = IMSLP_REGEXP_PAGE_COUNT.search(t.parent.text)
        if m is not None:
            try:
                page_count = int(m.group(1))
            except ValueError:
                pass

        file_id = int(t.text.replace("#", ""))

        t = s.find(attrs={"href": "/wiki/Special:GetFCtrStats/@{}".format(file_id)})
        if t is None:
            continue

        file_counter = int(t.text)

        # Fix image URL
        if f.imageinfo["url"][0] == "/":
            # URL is //imslp.org/stuff...
            f.imageinfo["url"] = "http:" + f.imageinfo["url"]

        images.append({
            "id": file_id,
            "rating": ratings_dict.get(file_id, dict()).get("rating", -1),
            "rating_count": ratings_dict.get(file_id, dict()).get("count", 0),
            "download_count": file_counter,
            "title": f_title,
            "url": f.imageinfo["url"],
            "page_count": page_count,
            "size": f.imageinfo.get("size"),
            "sha1": f.imageinfo.get("sha1"),
            "obj": f,
        })

    return images



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['imslp', 'imslp.helpers', 'imslp.interfaces']

package_data = \
{'': ['*']}

install_requires = \
['bs4>=0.0.1,<0.0.2', 'mwclient>=0.10.1,<0.11.0', 'requests>=2.24.0,<3.0.0']

setup_kwargs = {
    'name': 'imslp',
    'version': '0.2.0',
    'description': 'The clean and modern way of accessing IMSLP data and scores programmatically.',
    'long_description': '# imslp\n\n![pytest](https://github.com/jlumbroso/imslp/workflows/pytest/badge.svg)\n [![codecov](https://codecov.io/gh/jlumbroso/imslp/branch/master/graph/badge.svg?token=GX52420WN4)](https://codecov.io/gh/jlumbroso/imslp)\n [![Documentation Status](https://readthedocs.org/projects/imslp/badge/?version=latest)](https://imslp.readthedocs.io/en/latest/?badge=latest)\n [![Downloads](https://pepy.tech/badge/imslp)](https://pepy.tech/project/imslp)\n [![Run on Repl.it](https://repl.it/badge/github/jlumbroso/imslp)](https://repl.it/github/jlumbroso/imslp)\n [![Stargazers](https://img.shields.io/github/stars/jlumbroso/imslp?style=social)](https://github.com/jlumbroso/imslp)\n\n🎼 The clean and modern way of accessing IMSLP data and scores programmatically. 🎶\n\n## Installation\n\nThe package is available on PyPi and can be installed using your favorite package\nmanager:\n\n```shell\npip install imslp\n```\n\n## Data Sources\n\nThis project attempts to use robust sources of data, that do not require web scraping of some sort:\n\n- **MediaWiki API.** IMSLP is [one of tens of thousands of websites](https://wikiapiary.com/wiki/IMSLP)\nbuilt on top of [MediaWiki](https://www.mediawiki.org/wiki/MediaWiki), the framework created for\n[Wikipedia.org](https://en.wikipedia.org/wiki/MediaWiki). As such, it can be accessed through\nthe [MediaWiki API](https://www.mediawiki.org/wiki/API:Main_page) for which, fortunately,\nthere exists a fantastic Python wrapper library called [`mwclient`](https://github.com/mwclient/mwclient).\n\n- **IMSLP API.** For convenience, the IMSLP built some *ad-hoc* scripts that can be used to get a\nlist of people and a list of works, in a variety of different formats, including JSON.\n\nIt also uses scraping to collect additional information (such as the number of pages in a score, the\nnumber of times a score was downloaded, or the user-provided ratings).\n\n### Some quirks of IMSLP\n\nWhile fortunately, as mentioned, IMSLP uses a widely used open-source Wiki platform, MediaWiki, it has a\nhandful of quirks. Such as:\n\n- Composers are stored as `Category`, for instance `Category:Scarlatti, Domenico`. For each composer,\nthere is usually three tabs: "Compositions", "Collaborations" and "Collections"; these are stored as\nseparate categories resulting from the concatenation of the composer and subtype, such as\n`Category:Scarlatti, Domenico/Collections`.\n\n- PDF files for sheet music are stored as "images"; unfortunately, for the time being, the scheme does\nnot appear in the URLs computed for the files. These need to be manually patched.\n\n- The `imslpdisclaimeraccepted` cookie must be set to `"yes"` for files to download properly (otherwise,\ndownloading any file will result in the disclaimer page). With `mwclient`, this can be specified on login.\n    ```python\n    cookies = {\n        "imslp_wikiLanguageSelectorLanguage": "en",\n        "imslpdisclaimeraccepted": "yes",\n    }\n    ```\n\n- Much of the metadata associated with images, such as the internal ID or the download counter, is stored\nseparately than the MediaWiki metadata. This makes scraping the rendered HTML page a necessary endeavour.\n\nFortunately all these quirks are handled by this package!\n\n## Related Projects\n\nHere are a handful of other related projects available on GitHub to access the IMSLP data programmatically:\n\n- [jjjake/imslp-scrape](https://github.com/): Last commit in May 2012 (32 commits), mix of Python and shell, scraping\nthe website for data (people, score links) with HTML parsing.\n\n- [FrankTheCodeMonkey/IMSLP-Scraper](https://github.com/FrankTheCodeMonkey/IMSLP-Scraper): Last commit in June 2020 \n(6 commits), Python, scraping the website for data and scores, with HTML parsing and Selenium.\n\n- [josefleventon/imslp-api](https://github.com/josefleventon/imslp-api): Last commit in May 2020 (17 commits),\nJavaScript, uses [IMSLP\'s custom API](https://imslp.org/wiki/IMSLP:API) to get the list of people and list of works\nprogrammatically through a web API query. \n\nMore recently, and in other languages:\n\n- [IMSLP Instrument Information Parsing Program](https://github.com/yoonlight/imslp): Last commit in July 2020\n(47 commits), uses scraping to extract instrumentation information. \n\n## Acknowledgements\n\nLet\'s be clear that all the heavy lifting is done by [`mwclient`](https://github.com/mwclient/mwclient)—and\nthe volunteers who uploaded and/or scanned and/or typeset the scores on IMSLP. \n\n## License\n\nThis project is licensed under the LGPLv3 license, with the understanding\nthat importing a Python modular is similar in spirit to dynamically linking\nagainst a library.\n\n- You can use the library `imslp` in any project, for any purpose, as long\n  as you provide some acknowledgement to this original project for use of\n  the library.\n\n- If you make improvements to `imslp`, you are required to make those\n  changes publicly available.\n  ',
    'author': 'Jérémie Lumbroso',
    'author_email': 'lumbroso@cs.princeton.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jlumbroso/imslp',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)




"""
The clean and modern way of accessing IMSLP data and scores programmatically.
"""

__version__ = "0.1.0"
__author__ = "Jérémie Lumbroso <lumbroso@cs.princeton.edu>"

version_info = tuple(int(v) if v.isdigit() else v for v in __version__.split('.'))

